
#    Access Server  for  MIDAS Data Base

global env
source [file join $env(MIDASBASE) TclHttpd tcl MIDASDomainHandler.tcl]       ;#  load default support procedures

global MIDAS_DataBaseAccessServer_Exists
set MIDAS_DataBaseAccessServer_Exists 1

global MIDAS_DataBaseAccessServer_Verbose
set MIDAS_DataBaseAccessServer_Verbose 0

global MIDAS_DataBaseAccessServer_Implementation
set MIDAS_DataBaseAccessServer_Implementation 0

global MIDAS_SOAPDomainHandler_Verbose
set MIDAS_SOAPDomainHandler_Verbose 0

if {![info exists env(DataBaseRoot)]} {set env(DataBaseRoot) [file join $env(MIDASBASE) DB]}


if {![file isdirectory $env(DataBaseRoot)]} then {
    puts "Warning: MIDAS DataBase root $env(DataBaseRoot) is invalid or does not exist"
    set $env(DataBaseRoot)  [file join $env(MIDASBASE) DB]
    if {[file isdirectory $env(DataBaseRoot)]} then {puts "Using $env(DataBaseRoot)"}
}


global MIDAS_DataBaseCode MIDAS_DataBaseError MIDAS_DataBaseInfo
set MIDAS_DataBaseCode 0
set MIDAS_DataBaseError 0
set MIDAS_DataBaseInfo "no error present"

# Use namespaces to isolate our methods
namespace eval urn:DataBaseAccessServer {

    proc ServerVerbose {opt} {
        do_ServerVerbose $opt
        return [rpcvar::rpcvar int 0]
    }

    proc ServerImplementation {opt} {
        do_ServerImplementation $opt
        return [rpcvar::rpcvar int 0]
    }

    proc GetLastError {args} {
        set vl [do_GetLastError]
        return [list [list ResponseCode int [lindex $vl 0]] [list Code int [lindex $vl 1]]  [list Reason string [lindex $vl 2]]]
    }

    proc GetInformation {args} {
        set vl [do_GetInformation]
        return [list [list ResponseCode int [lindex $vl 0]] [list Reason string [lindex $vl 1]]]
    }

    proc do_ServerVerbose {option} {
        global MIDAS_DataBaseAccessServer_Verbose

        set MIDAS_DataBaseAccessServer_Verbose $option
    }

    proc do_ServerImplementation {option} {
        global MIDAS_DataBaseAccessServer_Implementation

        set MIDAS_DataBaseAccessServer_Implementation $option
    }

    proc do_GetLastError {} {
       global MIDAS_DataBaseCode MIDAS_DataBaseError MIDAS_DataBaseInfo

       return [list 0 $MIDAS_DataBaseError $MIDAS_DataBaseInfo]
    }

    proc do_GetInformation {} {
       global MIDAS_DataBaseCode MIDAS_DataBaseError MIDAS_DataBaseInfo

       return [list 0 "No further informaton available at present"]
    }

    proc DBnodes {nodeName pattern} {
       global MIDAS_DataBaseCode MIDAS_DataBaseError MIDAS_DataBaseInfo

        set result [do_nodes $nodeName $pattern]
        return [list [list ResponseCode int $MIDAS_DataBaseCode] [list result string $result]]
    }

    proc DBget {nodeName valueName} {
       global MIDAS_DataBaseCode MIDAS_DataBaseError MIDAS_DataBaseInfo

        set result [do_get $nodeName $valueName]
        return [list [list ResponseCode int $MIDAS_DataBaseCode] [list result string $result]]
    }

    proc DBset {nodeName valueName data type} {
       global MIDAS_DataBaseCode MIDAS_DataBaseError MIDAS_DataBaseInfo

        do_set $nodeName $valueName $data $type
        return [rpcvar::rpcvar int $MIDAS_DataBaseCode]
    }

    proc DBdelete {nodeName valueName} {
       global MIDAS_DataBaseCode MIDAS_DataBaseError MIDAS_DataBaseInfo

        do_delete $nodeName $valueName
        return [list [list ResponseCode int $MIDAS_DataBaseCode]]
    }

    proc DBtype {nodeName valueName} {
       global MIDAS_DataBaseCode MIDAS_DataBaseError MIDAS_DataBaseInfo

        set result [do_type $nodeName $valueName]
        return [list [list ResponseCode int $MIDAS_DataBaseCode] [list result string $result]]
    }

    proc DBvalues {nodeName pattern} {
       global MIDAS_DataBaseCode MIDAS_DataBaseError MIDAS_DataBaseInfo

        set result [do_values $nodeName $pattern]
        return [list [list ResponseCode int $MIDAS_DataBaseCode] [list result string $result]]
    }


    # We have to publish the public methods...

     SOAP::export ServerVerbose
     SOAP::export ServerImplementation
     SOAP::export nodes
     SOAP::export get
     SOAP::export set
     SOAP::export delete
     SOAP::export type
     SOAP::export values
     SOAP::export GetLastError
     SOAP::export GetInformation


    ##########    private procedures

    proc do_nodes {nodeName pattern} {                                 ;##########################
        global env MIDAS_DataBaseAccessServer_Implementation MIDAS_DataBaseAccessServer_Verbose
        global MIDAS_DataBaseCode MIDAS_DataBaseError MIDAS_DataBaseInfo

        if {$MIDAS_DataBaseAccessServer_Verbose} {puts "called do_nodes with args $nodeName; $pattern"}

        set  MIDAS_DataBaseCode 0
        set result {}

        set Separator "\\/"

#    first split the nodeName into components

        set NM [split $nodeName $Separator]

#    then join to make a file name

        set PATH $env(DataBaseRoot)
        foreach node $NM {set PATH [file join $PATH $node]}

#    check the nodeName is a directory

        if {[file isdirectory $PATH] == 0} {

           if {$MIDAS_DataBaseAccessServer_Implementation == 0} {
               set MIDAS_DataBaseError 0x10004
               set MIDAS_DataBaseInfo "node is invalid or does not exist"
               set MIDAS_DataBaseCode -1
               return $result
           } else {
               return -code -errorcode 0x10004 -errorinfo "node is invalid or does not exist"
           }
        }

#    and obtain directories contained

        if {$pattern == "" || $pattern == "NULL"} then {
            set z [catch {glob -directory $PATH -type d -- *} r]
        } else {
            set z [catch {glob -directory $PATH -type d -- $pattern} r]
        }

#    just return the subnode names but check for no directories

        if {$z == 1 && [string compare -length 16 "no files matched" $r] == 0} {
               set MIDAS_DataBaseError 0
               set MIDAS_DataBaseInfo "$r"
               set MIDAS_DataBaseCode 1
               return $result
        }

        if {$z != 0} {
               set MIDAS_DataBaseError 0x10004
               set MIDAS_DataBaseInfo "$r"
               set MIDAS_DataBaseCode -1
               return $result
        }

        foreach subnode $r {lappend result [file tail $subnode]}

        return $result
    }

    proc do_get {nodeName valueName} {                                 ;##########################
        global env MIDAS_DataBaseAccessServer_Implementation
        global MIDAS_DataBaseCode MIDAS_DataBaseError MIDAS_DataBaseInfo

        set  MIDAS_DataBaseCode 0
        set result {}

        set Separator "\\/"

#    first split the nodeName into components

        set NM [split $nodeName $Separator]

#    then join to make a file name

        set PATH $env(DataBaseRoot)
        foreach node $NM {set PATH [file join $PATH $node]}

#    check the nodeName is a directory

        if {[file isdirectory $PATH] == 0} {
           if {$MIDAS_DataBaseAccessServer_Implementation == 0} {
               set MIDAS_DataBaseError 0x10004
               set MIDAS_DataBaseInfo "node is invalid or does not exist"
               set MIDAS_DataBaseCode -1
               return $result
           } else {
               return -code -errorcode 0x10004 -errorinfo "node is invalid or does not exist"
           }
        }

#    read the CONTENTS and put into an array

        set z [catch {open [file join $PATH CONTENTS] r} f]
        if {$z != 0} {
           if {$MIDAS_DataBaseAccessServer_Implementation == 0} {
               set MIDAS_DataBaseError 0x10004
               set MIDAS_DataBaseInfo "unable to access CONTENTS for node $PATH: $f"
               set MIDAS_DataBaseCode -1
               return $result
           } else {
               return -code -errorcode 0x10004 -errorinfo "unable to access CONTENTS for node $PATH: $f"
           }
        }
        set contents "[read $f]"
        close $f

        catch {unset Contents}

        foreach item [split $contents \n] {
            if {$item == ""} {continue}
            set Contents([head $item]) [tail $item]
        }

#    obtain the value

        if {[info exists Contents($valueName)] == 0} {
           if {$MIDAS_DataBaseAccessServer_Implementation == 0} {
               set MIDAS_DataBaseError 0x10004
               set MIDAS_DataBaseInfo "value $valueName does not exist in node $PATH"
               set MIDAS_DataBaseCode -1
               return $result
           } else {
               return -code -errorcode 0x10004 -errorinfo "value $valueName does not exist in node $PATH"
           }
        }

        set vl $Contents($valueName)

        return [tail $vl]
    }

    proc do_set {nodeName valueName data type} {                               ;##########################
        global env MIDAS_DataBaseAccessServer_Implementation
        global MIDAS_DataBaseCode MIDAS_DataBaseError MIDAS_DataBaseInfo

        set  MIDAS_DataBaseCode 0

        set Separator "\\/"

#    first split the nodeName into components

        set NM [split $nodeName $Separator]

#    then join to make a file name

        set PATH $env(DataBaseRoot)
        foreach node $NM {set PATH [file join $PATH $node]}

#    check nodeName is valid and create nodeName if necessary

        if {[file exists $PATH] == 0} {

#    nodeName does not exist  - create it

            set z [catch {file mkdir $PATH} m]
            if {$z != 0} {

                if {$MIDAS_DataBaseAccessServer_Implementation == 0} {
                   set MIDAS_DataBaseError 0x10004
                   set MIDAS_DataBaseInfo "unable to create node $PATH: $m"
                   set MIDAS_DataBaseCode -1
                   return 
               } else {
                   return -code -errorcode 0x10004 -errorinfo "unable to create node $PATH: $m"
               }
            }
        } else {

#    nodeName exists - check it is a directory

            if {[file isdirectory $PATH] == 0} {

               if {$MIDAS_DataBaseAccessServer_Implementation == 0} {
                   set MIDAS_DataBaseError 0x10004
                   set MIDAS_DataBaseInfo "nodeName is invalid"
                   set MIDAS_DataBaseCode -1
                   return 
               } else {
                   return -code -errorcode 0x10004 -errorinfo "nodeName is invalid"
               }
            }
        }

#    check for the node CONTENTS file and create it if needed

        set z [catch {open [file join $PATH CONTENTS] a} f]
        if {$z != 0} {

               if {$MIDAS_DataBaseAccessServer_Implementation == 0} {
                   set MIDAS_DataBaseError 0x10004
                   set MIDAS_DataBaseInfo "unable to access CONTENTS for node $PATH: $f"
                   set MIDAS_DataBaseCode -1
                   return 
               } else {
                   return -code -errorcode 0x10004 -errorinfo "unable to access CONTENTS for node $PATH: $f"
               }
        }
        close $f

        if {$valueName == "" || $valueName == "NULL"} then {return 0}    ;#  just create the node

#    read the CONTENTS and put into an array

        set z [catch {open [file join $PATH CONTENTS] r} f]
        if {$z != 0} {

               if {$MIDAS_DataBaseAccessServer_Implementation == 0} {
                   set MIDAS_DataBaseError 0x10004
                   set MIDAS_DataBaseInfo "unable to access CONTENTS for node $PATH: $f"
                   set MIDAS_DataBaseCode -1
                   return 
               } else {
                   return -code -errorcode 0x10004 -errorinfo "unable to access CONTENTS for node $PATH: $f"
               }
        }
        set contents "[read $f]"
        close $f

        catch {unset Contents}

        foreach item [split $contents \n] {
            if {$item == ""} {continue}
            set Contents([head $item]) [tail $item]
        }

#    create or replace the item valueName

        switch $type {
            float -
            string -
            long -
            short -
            binary   {set T $type}
            default  {set T string}
        }

        set Contents($valueName) "$T $data"

#    finally rewrite the CONTENTS

        set z [catch {open [file join $PATH CONTENTS] w} f]
        if {$z != 0} {

               if {$MIDAS_DataBaseAccessServer_Implementation == 0} {
                   set MIDAS_DataBaseError 0x10004
                   set MIDAS_DataBaseInfo "unable to access CONTENTS for node $PATH: $f"
                   set MIDAS_DataBaseCode -1
                   return 1
               } else {
                   return -code -errorcode 0x10004 -errorinfo "unable to access CONTENTS for node $PATH: $f"
               }
        }

        foreach item [array names Contents] {
             puts $f  "$item $Contents($item)"
        }

        close $f

        return 
    }

     proc do_delete {nodeName valueName} {                                ;##########################
        global env MIDAS_DataBaseAccessServer_Implementation
        global MIDAS_DataBaseCode MIDAS_DataBaseError MIDAS_DataBaseInfo

        set  MIDAS_DataBaseCode 0

        set Separator "\\/"

#    first split the nodeName into components

        set NM [split $nodeName $Separator]

#    then join to make a file name

        set PATH $env(DataBaseRoot)
        foreach node $NM {set PATH [file join $PATH $node]}

#    check the nodeName is a directory

        if {[file isdirectory $PATH] == 0} {

               if {$MIDAS_DataBaseAccessServer_Implementation == 0} {
                   set MIDAS_DataBaseError 0x10004
                   set MIDAS_DataBaseInfo "node is invalid or does not exist"
                   set MIDAS_DataBaseCode -1
                   return 
               } else {
                   return -code -errorcode 0x10004 -errorinfo "node is invalid or does not exist"
               }
        }

        if {$valueName == "" || $valueName == "NULL"} {

#    delete the node and all subnodes

            set z [catch {file delete -force $PATH} m]
            if {$z != 0} {

               if {$MIDAS_DataBaseAccessServer_Implementation == 0} {
                   set MIDAS_DataBaseError 0x10004
                   set MIDAS_DataBaseInfo "unable to delete node $PATH: $m"
                   set MIDAS_DataBaseCode -1
                   return 
               } else {
                   return -code -errorcode 0x10004 -errorinfo "unable to delete node $PATH: $m"
               }
            }
        } else {

#    just delete the value

#    read the CONTENTS and put into an array

           set z [catch {open [file join $PATH CONTENTS] r} f]
           if {$z != 0} {

               if {$MIDAS_DataBaseAccessServer_Implementation == 0} {
                   set MIDAS_DataBaseError 0x10004
                   set MIDAS_DataBaseInfo "unable to access CONTENTS for node $PATH: $f"
                   set MIDAS_DataBaseCode -1
                   return 
               } else {
                   return -code -errorcode 0x10004 -errorinfo "unable to access CONTENTS for node $PATH: $f"
               }
           }
           set contents "[read $f]"
           close $f

           catch {unset Contents}
   
           foreach item [split $contents \n] {
               if {$item == ""} {continue}
               set Contents([head $item]) [tail $item]
           }

           catch {unset Contents($valueName)}    ;# remove the value

#    finally rewrite the CONTENTS

           set z [catch {open [file join $PATH CONTENTS] w} f]
           if {$z != 0} {

               if {$MIDAS_DataBaseAccessServer_Implementation == 0} {
                   set MIDAS_DataBaseError 0x10004
                   set MIDAS_DataBaseInfo "unable to access CONTENTS for node $PATH: $f"
                   set MIDAS_DataBaseCode -1
                   return 
               } else {
                   return -code -errorcode 0x10004 -errorinfo "unable to access CONTENTS for node $PATH: $f"
               }
           }

           foreach item [array names Contents] {
                puts $f  "$item $Contents($item)"
           }

           close $f
        }

        return 
    }


    proc do_type {nodeName valueName} {                                ;##########################
        global env MIDAS_DataBaseAccessServer_Implementation
        global MIDAS_DataBaseCode MIDAS_DataBaseError MIDAS_DataBaseInfo

        set  MIDAS_DataBaseCode 0
        set result {}

        set Separator "\\/"

#    first split the nodeName into components

        set NM [split $nodeName $Separator]

#    then join to make a file name

        set PATH $env(DataBaseRoot)
        foreach node $NM {set PATH [file join $PATH $node]}

#    check the nodeName is a directory

        if {[file isdirectory $PATH] == 0} {

               if {$MIDAS_DataBaseAccessServer_Implementation == 0} {
                   set MIDAS_DataBaseError 0x10004
                   set MIDAS_DataBaseInfo "node is invalid or does not exist"
                   set MIDAS_DataBaseCode -1
                   return $result
               } else {
                   return -code -errorcode 0x10004 -errorinfo "node is invalid or does not exist"
               }
        }

#    read the CONTENTS and put into an array

        set z [catch {open [file join $PATH CONTENTS] r} f]
        if {$z != 0} {

               if {$MIDAS_DataBaseAccessServer_Implementation == 0} {
                   set MIDAS_DataBaseError 0x10004
                   set MIDAS_DataBaseInfo "unable to access CONTENTS for node $PATH: $f"
                   set MIDAS_DataBaseCode -1
                   return $result
               } else {
                   return -code -errorcode 0x10004 -errorinfo "unable to access CONTENTS for node $PATH: $f"
               }
        }
        set contents "[read $f]"
        close $f

        catch {unset Contents}

        foreach item [split $contents \n] {
            if {$item == ""} {continue}
            set Contents([head $item]) [tail $item]
        }

#    obtain the type

        if {[info exists Contents($valueName)] == 0} {

               if {$MIDAS_DataBaseAccessServer_Implementation == 0} {
                   set MIDAS_DataBaseError 0x10004
                   set MIDAS_DataBaseInfo "value $valueName does not exist in node $PATH"
                   set MIDAS_DataBaseCode -1
                   return $result
               } else {
                   return -code -errorcode 0x10004 -errorinfo "value $valueName does not exist in node $PATH"
               }
        }

        set vl $Contents($valueName)

        return [head $vl]
    }


    proc do_values {nodeName pattern} {                                ;##########################
        global env MIDAS_DataBaseAccessServer_Implementation
        global MIDAS_DataBaseCode MIDAS_DataBaseError MIDAS_DataBaseInfo

        set  MIDAS_DataBaseCode 0
        set result {}

        set Separator "\\/"

#    first split the nodeName into components

        set NM [split $nodeName $Separator]

#    then join to make a file name

        set PATH $env(DataBaseRoot)
        foreach node $NM {set PATH [file join $PATH $node]}

#    check the nodeName is a directory

        if {[file isdirectory $PATH] == 0} {

               if {$MIDAS_DataBaseAccessServer_Implementation == 0} {
                   set MIDAS_DataBaseError 0x10004
                   set MIDAS_DataBaseInfo "node is invalid or does not exist"
                   set MIDAS_DataBaseCode -1
                   return $result
               } else {
                   return -code -errorcode 0x10004 -errorinfo "node is invalid or does not exist"
               }
        }

#    read the CONTENTS and put into an array

        set z [catch {open [file join $PATH CONTENTS] r} f]
        if {$z != 0} {

               if {$MIDAS_DataBaseAccessServer_Implementation == 0} {
                   set MIDAS_DataBaseError 0x10004
                   set MIDAS_DataBaseInfo "unable to access CONTENTS for node $PATH: $f"
                   set MIDAS_DataBaseCode -1
                   return $result
               } else {
                   return -code -errorcode 0x10004 -errorinfo "unable to access CONTENTS for node $PATH: $f"
               }
        }
        set contents "[read $f]"
        close $f

        catch {unset Contents}

        foreach item [split $contents \n] {
            if {$item == ""} {continue}
            set Contents([head $item]) [tail $item]
        }

#    return all matching values

        if {$pattern == "" || $pattern == "NULL"} then {set pat *} else {set pat $pattern}

        foreach item [array names Contents] {
             if {[string match $pat $item] == 1} then {lappend result $item}
        }

        return $result
    }

     
}   ;#  end of    namespace eval urn:DataBaseAccessServer


# register this service with tclhttpd
puts [SOAP::Domain::register  -prefix  /DataBaseAccessServer -namespace urn:DataBaseAccessServer -uri urn:DataBaseAccessServer]


proc ::SOAP::Domain::MIDAS_SOAPDomainHandler_/DataBaseAccessServer {optsname sock args} {   \
    ::SOAP::Domain::MIDAS_SOAPDomainHandler_/DefaultAccessServer $optsname $sock $args}



